import logging

from wokkel import client, xmppim, subprotocols
from twisted.internet import reactor, ssl
from twisted.words.protocols.jabber.jid import JID
from twisted.words.xish import domish
from twisted.application import internet

import ibid
from ibid.config import Option, IntOption, BoolOption, ListOption
from ibid.source import IbidSourceFactory
from ibid.event import Event

class Message(domish.Element):

    def __init__(self, to, frm, body, type):
        domish.Element(self, (None, 'message'))
        self['to'] = to
        self['from'] = frm
        self['type'] = 'chat'
        self.addElement('body', content=body)

class JabberBot(xmppim.MessageProtocol, xmppim.PresenceClientProtocol, xmppim.RosterClientProtocol):

    def __init__(self):
        xmppim.MessageProtocol.__init__(self)
        self.rooms = []

    def connectionInitialized(self):
        self.parent.log.info(u"Connected")
        xmppim.MessageProtocol.connectionInitialized(self)
        xmppim.PresenceClientProtocol.connectionInitialized(self)
        xmppim.RosterClientProtocol.connectionInitialized(self)
        self.xmlstream.send(xmppim.AvailablePresence())
        self.roster = self.getRoster() #See section 7.3 of http://www.ietf.org/rfc/rfc3921.txt 
        self.name = self.parent.name
        self.parent.send = self.send
        self.parent.proto = self
        for room in self.parent.rooms:
            self.join(room)

        event = Event(self.parent.name, u'source')
        event.status = u'connected'
        ibid.dispatcher.dispatch(event)

    def connectionLost(self, reason):
        self.parent.log.info(u"Disconnected (%s)", reason)

        event = Event(self.parent.name, u'source')
        event.status = u'disconnected'
        ibid.dispatcher.dispatch(event)

        subprotocols.XMPPHandler.connectionLost(self, reason)

    def _state_event(self, entity, state):
        event = Event(self.name, u'state')
        event.state = state
        if entity.userhost().lower() in self.rooms:
            nick = entity.full().split('/')[1]
            event.channel = entity.userhost()
            if nick == self.parent.nick:
                event.type = u'connection'
                event.status = state == u'online' and u'joined' or u'left'
            else:
                event.sender['connection'] = entity.full()
                event.sender['id'] = event.sender['connection']
                event.sender['nick'] = nick
                event.public = True
        else:
            event.sender['connection'] = entity.full()
            event.sender['id'] = event.sender['connection'].split('/')[0]
            event.sender['nick'] = event.sender['connection'].split('@')[0]
            event.channel = entity.full()
            event.public = False
        ibid.dispatcher.dispatch(event).addCallback(self.respond)

    def availableReceived(self, entity, show=None, statuses=None, priority=0):
        self.parent.log.debug(u"Received available presence from %s (%s)", entity.full(), show)
        self._state_event(entity, u'online')

    def unavailableReceived(self, entity, statuses):
        self.parent.log.debug(u"Received unavailable presence from %s", entity.full())
        self._state_event(entity, u'offline')

    def subscribeReceived(self, entity):
        response = xmppim.Presence(to=entity, type='subscribed')
        self.xmlstream.send(response)
        response = xmppim.Presence(to=entity, type='subscribe')
        self.xmlstream.send(response)
        self.parent.log.info(u"Received and accepted subscription request from %s", entity.full())

    def onMessage(self, message):
        self.parent.log.debug(u"Received %s message from %s: %s", message['type'], message['from'], message.body)

        if message.x and message.x.defaultUri == 'jabber:x:delay':
            self.parent.log.debug(u"Ignoring delayed message")
            return

        if self.parent.accept_domains:
            if message['from'].split('/')[0].split('@')[1] not in self.parent.accept_domains:
                self.parent.log.info(u"Ignoring message because sender is not in accept_domains")
                return

        if message.body is None:
            self.parent.log.info(u'Ignoring empty message')
            return

        event = Event(self.parent.name, u'message')
        event.message = unicode(message.body)
        event.sender['connection'] = message['from']

        if message['type'] == 'groupchat':
            event.sender['id'] = message['from'].find('/') != -1 and message['from'].split('/')[1] or message['from']
            if event.sender['id'] == self.parent.nick:
                return
            event.sender['nick'] = event.sender['id']
            event.channel = message['from'].split('/')[0]
            event.public = True
        else:
            event.sender['id'] = event.sender['connection'].split('/')[0]
            event.sender['nick'] = event.sender['connection'].split('@')[0]
            event.channel = event.sender['connection']
            event.public = False
            event.addressed = True

        ibid.dispatcher.dispatch(event).addCallback(self.respond)

    def respond(self, event):
        for response in event.responses:
            self.send(response)

    def send(self, response):
        message = domish.Element((None, 'message'))
        message['to'] = response['target']
        message['from'] = self.parent.authenticator.jid.full()
        if message['to'] in self.rooms:
            message['type'] = 'groupchat'
        else:
            message['type'] = 'chat'
        message.addElement('body', content=response['reply'])
        self.xmlstream.send(message)
        self.parent.log.debug(u"Sent %s message to %s: %s", message['type'], message['to'], message.body)

    def join(self, room):
        self.parent.log.info(u"Joining %s", room)
        jid = JID('%s/%s' % (room, self.parent.nick))
        presence = xmppim.AvailablePresence(to=jid)
        self.xmlstream.send(presence)
        self.rooms.append(room.lower())

    def leave(self, room):
        self.parent.log.info(u"Leaving %s", room)
        jid = JID('%s/%s' % (room, self.parent.nick))
        presence = xmppim.UnavailablePresence(to=jid)
        self.xmlstream.send(presence)
        self.rooms.remove(room.lower())

class SourceFactory(client.DeferredClientFactory, IbidSourceFactory):

    auth = ('implicit',)
    supports = ('multiline',)

    port = IntOption('port', 'Server port number')
    ssl = BoolOption('ssl', 'Usel SSL', False)
    server = Option('server', 'Server hostname')
    jid_str = Option('jid', 'Jabber ID')
    password = Option('password', 'Jabber password')
    nick = Option('nick', 'Nick for chatrooms', ibid.config['botname'])
    rooms = ListOption('rooms', 'Chatrooms to autojoin', [])
    accept_domains = Option('accept_domains', 'Only accept messages from these domains')

    def __init__(self, name):
        IbidSourceFactory.__init__(self, name)
        self.log = logging.getLogger('source.%s' % name)
        client.DeferredClientFactory.__init__(self, JID(self.jid_str), self.password)
        bot = JabberBot()
        self.addHandler(bot)
        bot.setHandlerParent(self)

    def setServiceParent(self, service):
        if self.ssl:
            sslctx = ssl.ClientContextFactory()
            port = self.port or 5223
            if service:
                internet.SSLClient(self.server, port, self, sslctx).setServiceParent(service)
            else:
                reactor.connectSSL(self.server, port, self, sslctx)
        else:
            port = self.port or 5222
            if service:
                internet.TCPClient(self.server, port, self).setServiceParent(service)
            else:
                reactor.connectTCP(self.server, port, self)

    def connect(self):
        return self.setServiceParent(None)

    def disconnect(self):
        self.stopFactory()
        self.proto.xmlstream.transport.loseConnection()
        return True

    def join(self, room):
        return self.proto.join(room)

    def leave(self, room):
        return self.proto.leave(room)

    def url(self):
        return u'xmpp://%s' % (self.jid_str,)

    def logging_name(self, identity):
        return identity.split('/')[0]

# vi: set et sta sw=4 ts=4:
